% optimal control computation in double well potential V
% cost is int_0^T 0.5 R u^2 + V(x)
% dynamics is dx=dxi + u dt
clear all

xmin=-2;
xmax=3;
a=0.25;
b=0.75;
c=0.5;

n=100;		% number of sample paths 

R=1;
T=10;
nu=1;
dt=0.01;
N=T/dt;		% number of time points
lambda=R*nu;


% initialize x
%rand('state',0);
%randn('state',0);
x0=xmin+(xmax-xmin)*rand;
x0=1;

V0=1*ones(1,N);
alive=ones(1,n);
w=ones(1,n);
xall=zeros(n,N);
METHOD='adaptive';
switch METHOD
case 'weighted'
	figure(1);
case 'stochastic'
	figure(2);
case 'adaptive'
	figure(3);
end

for i=1:n,
	x=x0;
	j=0;	
	while(j<N)&(alive(i))
		j=j+1;
		x=x+sqrt(nu*dt)*randn;
		Vx=V(x,a,b,c);
		switch METHOD
		case 'weighted'
			w(i)=w(i)*exp(-Vx*dt/lambda);
		case 'stochastic'
			if (rand<(Vx-V0(j))*dt/lambda)
				alive(i)=0;
			else
				w(i)=w(i)*exp(-(Vx-V0(j))*dt/lambda);
			end;
		case 'adaptive'
			if (rand<(Vx-V0(j))*dt/lambda)
				alive(i)=0;
			else
				w(i)=w(i)*exp(-(Vx-V0(j))*dt/lambda);
			end;
			V0(j)=min(V0(j),Vx);
		end;
		xall(i,j)=x;
	end;
end;
sum(alive)
xallalive=xall(find(alive),:);
walive=w(find(alive));
subplot(2,2,1)
dxd=0.01;
xgrid=xmin:dxd:xmax;
plot(xgrid,V(xgrid,a,b,c))
subplot(2,2,2)
plot(xallalive')	
subplot(2,2,3)
plot(x0,0,'*r')
hold on
plot(xallalive(:,N),walive/N,'*')
hold off
	



