% optimal control computation in mountain car with mountain L
% cost is int_0^T 0.5 R u^2 + V(x)
% dynamics is dx=v dt
% 				dv= F_g dt + u dt + d\xi
clear all

SAMPLING='weighted';	% weighted, stochastic, adaptive'
LEARNINGFG='off';		% on, off
METHOD='J';			% dxi, J

xmin=-4;
xmax=4;
vmin=-2;
vmax=2;
dxd=0.01;
a=0.5;
b=0.75;
c=0.5;
a=0;
b=1;
c=0;


n=50;		% number of sample paths 

R=1;
Ttotal=20;
T=2;			% horizon time
nu=2;
G=10;		% gravitational constant
dt=0.01;
N=Ttotal/dt;		% number of time points
N1=N-T/dt;
lambda=R*nu;


% initialize car
rand('state',0);
randn('state',0);
x0=0.2;
v0=0;

V0=1*ones(1,N);
alive=ones(1,n);
w=zeros(n,N);
xall=zeros(n,N);
vall=zeros(n,N);


x=x0;
v=v0;
xnoiseless=zeros(1,N);
vnoiseless=zeros(1,N);
for j=1:N,
	Vx=V(x,a,b,c);
	Vx1=V(x+dxd,a,b,c);
	dV=(Vx1-Vx)/dxd;
	f=G*dV/sqrt(1+dV^2);
	v=v+f*dt/2;
	x=x+v*dt;
	v=v+f*dt/2;
	xnoiseless(j)=x;
	vnoiseless(j)=v;
end;
figure(1)
clf
subplot(3,2,2)
plot((1:N)*dt,xnoiseless);
axis([0 Ttotal xmin xmax]);
xlabel('t');
ylabel('x');
	
xiall=sqrt(nu*dt)*randn(n,N-1);
for i=1:n,
	x=x0;
	v=v0;
	for j=1:N-1,
	if (x<xmax)&(x>xmin)
		Vx=V(x,a,b,c);
		Vx1=V(x+dxd,a,b,c);
		dV=(Vx1-Vx)/dxd;
		f=G*dV/sqrt(1+dV^2);
		x=x+v*dt;
		v=v+f*dt+xiall(i,j);
		switch SAMPLING
		case 'weighted'
			w(i,j)=-Vx*dt/lambda;
		case 'stochastic'
			if (rand<(Vx-V0(j))*dt/lambda)
				alive(i)=0;
				xend(i)=x;
				jend(i)=j;
			else
				w(i,j)=-(Vx-V0(j))*dt/lambda;
			end;
		case 'adaptive'
			if (rand<(Vx-V0(j))*dt/lambda)
				alive(i)=0;
				xend(i)=x;
				jend(i)=j;
			else
				w(i,j)=-(Vx-V0(j))*dt/lambda;
			end;
			V0(j)=min(V0(j),Vx);
		end;
	end;
	xall(i,j)=x;
	vall(i,j)=v;
	end;
end;


subplot(3,2,1)
xgrid=xmin:dxd:xmax;
hold on
plot(xgrid,-V(xgrid,a,b,c))
plot(x0,-V(0,a,b,c),'*r')
xlabel('x')
ylabel('L=-V')
hold off
subplot(3,2,3)
hold on
for i=1:n,
	plot((1:N)*dt,xall(i,:))
end;
axis([0 Ttotal xmin xmax])
xlabel('t')
ylabel('x')


% compute psi^mu X^mu dX^mu
logpsi=zeros(n,N1);
for j=1:N1
    logpsi(:,j)=sum(w(:,j:j+T/dt-1),2);
end;
xmu=reshape(xall(:,1:N1),1,n*N1);
vmu=reshape(vall(:,1:N1),1,n*N1);
ximu=reshape(xiall(:,1:N1),1,n*N1);
logpsimu=reshape(logpsi,1,n*N1);



% controlled trajectory

x=x0;
v=v0;
Xall=zeros(2,N1);
Uall=zeros(1,N1);
Vall=zeros(1,N1);
lmuset=zeros(1,N1);
meand=zeros(1,N1);
stdd=zeros(1,N1);
figure(2)
for j=1:N1,
	X=[x;v];
	Xall(:,j)=X;
if (x<xmax)&(x>xmin)

	switch METHOD
	case 'J',
		m=10;	% neighborhoodsize
		dv=0.1;
		v1=v-dv;
		dmu=sqrt((x-xmu).^2+(v1-vmu).^2);
		% find m nearest X^mu
		[dmusort dmui]=sort(dmu);
		muset=dmui(1:m);
		xmuset1=xmu(muset);
		vmuset1=vmu(muset);
    	logpsimuset1=logpsimu(muset);
    	J1=-lambda*log(mean(exp(logpsimuset1)));
    	J1=-lambda*max(logpsimuset1);
		v1=v+dv;
		dmu=sqrt((x-xmu).^2+(v1-vmu).^2);
		% find m nearest X^mu
		[dmusort dmui]=sort(dmu);
		muset=dmui(1:m);
		xmuset2=xmu(muset);
		vmuset2=vmu(muset);
    	logpsimuset2=logpsimu(muset);
    	J2=-lambda*log(mean(exp(logpsimuset2)));
    	J2=-lambda*max(logpsimuset2);
		u=-1/R*(J2-J1)/dv/2;
	case 'dxi',
		% find m nearest X^mu
		dmu=sqrt((x-xmu).^2+(v-vmu).^2);
		if 0
			m=1000;	% neighborhoodsize
			[dmusort dmui]=sort(dmu);
			muset=dmui(1:m);
		else
			dv=0.1;
			muset=find(dmu<dv);
		end;
		lmuset(j)=length(muset);
		meand(j)=mean(dmu(muset));
		stdd(j)=std(dmu(muset));
		xmuset=xmu(muset);
		vmuset=vmu(muset);
		ximuset=ximu(muset);
    	psimuset=exp(logpsimu(muset));
        u=psimuset*ximuset'/sum(psimuset)/dt;
	end;
		

	switch LEARNINGFG
	case 'on'
		% compute f(x)
		etaf=1/m*DXmu(:,muset)*deltaXmu'/dt;
		barf=mean(DXmu(:,muset),2)/dt;
		%std(DXmu(:,muset),[],2)/dt
		wf=etaf*inv(chi);
		thetaf=barf-wf*barX;
		fest=wf*X+thetaf;
		
		% compute g(x)
		klad=DXmu(:,muset)-fest*ones(1,m)*dt;
		sigma2=1/m*klad*klad';
		[U D]=eig(sigma2);
		if U(1,1)<0, U(:,1)=-U(:,1);end;
		if U(2,2)<0, U(:,2)=-U(:,2);end;
		gest=U*sqrt(D)/sqrt(dt);
	
		% compute u(x)
		u=nu*gest'*wpsi/psi
		Vx=V(x,a,b,c);
		Vx1=V(x+dxd,a,b,c);
		dV=(Vx1-Vx)/dxd;
		f=G*dV/sqrt(1+dV^2);
    	x=x+v*dt;
    	v=v+f*dt+u*dt+sqrt(nu*dt)*randn;
	case 'off'
		if (j==20)
			clf
			hold on
			scatter(xmu,vmu,3,-lambda*logpsimu)
			colorbar
			plot(x,v,'*k',[x x+v*dt],[v v+(f+u)*dt],'k')
			switch METHOD
			case 'J'
				plot(xmuset1,vmuset1,'.k',xmuset2,vmuset2,'k')
			case 'dxi'
				plot(xmuset,vmuset,'.k');
			end;
			hold off
			drawnow
		end;
		Vx=V(x,a,b,c);
		Vx1=V(x+dxd,a,b,c);
		dV=(Vx1-Vx)/dxd;
		f=G*dV/sqrt(1+dV^2);
    	x=x+v*dt;
    	v=v+(f+u)*dt;
	end;	
	Vall(j)=Vx;
	Uall(j)=u;
	%pause
end;
end;
%fprintf('total realized cost = %f\n',sum(Vall(1:j))*dt+0.5*R*sum(uall(1:j).^2)*dt);
%fprintf('total time = %f\n',j*dt);
figure(1)
subplot(3,2,3)
hold on
plot((1:N1)*dt,Xall(1,:),'k','LineWidth',2)
axis([0 Ttotal xmin xmax])
hold off
subplot(3,2,4)
plot((1:N1)*dt,Uall,(1:N1)*dt,Xall(2,:),'k')
xlabel('t');
ylabel('u');
subplot(3,2,5)
plot((1:N1)*dt,lmuset);
xlabel('t');
ylabel('# neighbors');
subplot(3,2,6)
plot((1:N1)*dt,meand,(1:N1)*dt,meand+stdd,(1:N1)*dt,meand-stdd);
xlabel('t');
ylabel('average distance');

