function y=plant(x,xi,dt,PLANT,W)

% given a state x and a control xi it outputs a state
% y=x+f(x) dt + g(x) xi	


switch PLANT,
case 'acrobot',
	% check dimensions
	if ~(length(x)==4) fprintf('ERROR: x0 has incorrect dimension %d\n', length(x)); end;
	klad=size(xi);
	if ~(klad(1)==1) fprintf('ERROR: xi has incorrect dimension %d %d\n', klad); end;
	c1=2.663;
	G=9.8;
	d11=c1+2*cos(x(2));
	d22=1.33;
	d12=d22+cos(x(2));
	h1=-sin(x(2))*(x(4)^2+2*x(3)*x(4));
	h2=sin(x(2))*x(3)^2;
	phi2=G*cos(x(1)+x(2));
	phi1=1.5*G*cos(x(1))+phi2;
	D=d11*d22-d12^2;
	f=zeros(1,4);
	f(1)=x(3);
	f(2)=x(4);
	f(3)=1/D*(-d22*(h1+phi1)+d12*(h2+phi2));
	f(4)=1/D*(d12*(h1+phi1)-d11*(h2+phi2));
	g=zeros(1,4);
	g(3)=-d12/D;
	g(4)=d11/D;
	y=x+f'*dt+g'*xi;
case 'singlewell',
	y=x + xi;
case 'neuralnetwork',
	y=(1-dt)*x+ W*tanh(x)*dt+xi;
case 'lorenz'
	rho=50; sigma=10; beta=8/3;
	dx=[sigma*(x(2)-x(1)),x(1)*(rho-x(3))-x(2), x(1)*x(2)-beta*x(3)]';
	y=(1-dt)*x+dx*dt+xi;
end;
