% optimal control computation in quadratic potential 
% cost is int_0^T 0.5 R u^2 + 0.5 Q x^2
% dynamics is dx=d\xi
clear all

xmin=-10;
xmax=10;
dxd=0.01;

n=100;		% number of sample paths 

R=1;
Q=5;
Ttotal=5;
T=1;		% horizon time
nu=1;
dt=0.05;
N=Ttotal/dt;		% number of time points
N1=N-T/dt;
lambda=R*nu;


% initialize 
%rand('state',1);
%randn('state',1);
x0=2+2*rand(1,n);
x0=xmin+(xmax-xmin)*rand(1,n);
x0=2;

xall=zeros(n,N);
METHOD='J';			% psi, dxmu, J
METHOD='dxmu';			% psi, dxmu, J

xall(:,1)=x0';
xiall=sqrt(nu*dt)*randn(n,N-1);
for j=1:N-1,
	xall(:,j+1)=xall(:,j)+xiall(:,j);
end;
w=-0.5*Q*xall.^2*dt/lambda;

switch METHOD
case 'dxmu'
	figure(1)
	clf
case 'J'
	figure(2)
	clf
end;

subplot(4,2,1)
hold on
for i=1:n,
	plot((1:N)*dt,xall(i,:),'r');
end;
axis([0 N*dt xmin xmax])
xlabel('t')
ylabel('x')
hold off
disp('all trajectories computed')

% compute psi^mu x^mu 
logpsi=zeros(n,N1);
for j=1:N1
	logpsi(:,j)=sum(w(:,j:j+T/dt-1),2);
end;
xmu=reshape(xall(:,1:N1),1,n*N1);
ximu=reshape(xiall(:,1:N1),1,n*N1);
logpsimu=reshape(logpsi,1,n*N1);
a=exp(2*sqrt(Q/R)*T);
P=R*sqrt(Q/R)*(a-1)/(1+a);
subplot(4,2,4)
xgrid=xmin:dxd:xmax;
plot(xmu,-lambda*logpsimu,'.',xgrid,0.5*P*xgrid.^2,'r')
legend('computed','exact');
axis([-5 5 0 20])
xlabel('x');
ylabel('J')



% controlled trajectory

x=x0;
xcontrol=zeros(1,N1);
uall=zeros(1,N1);
Vall=zeros(1,N1);
deltax=zeros(1,N1);
meanx=zeros(1,N1);
dx=0.5;
m=500;
for j=1:N1,
	xcontrol(j)=x;

	% compute u(x)
	switch METHOD
	case 'psi'
		% find m nearest x^mu
		dmu=sqrt((x-xmu).^2);
		[dmusort dmui]=sort(dmu);
		muset=dmui(1:m);
		muset=find(dmu<100*sqrt(nu)*dt);
		m=length(muset)
		psimuset=psimu(muset);
		xmuset=xmu(muset);
	
		% compute psi(x) and nabla psi(x)
		barpsi=mean(psimuset);
		barx=mean(xmuset);
		deltaxmu=xmuset-barx;
		etapsi=1/m*deltaxmu*psimuset';
		chi=1/m*deltaxmu*deltaxmu';
		wpsi=inv(chi)*etapsi;
		thetapsi=barpsi-wpsi'*barx;
		psi=wpsi'*x+thetapsi;

		u=nu*wpsi/psi;
		xgrid=min(xmuset):0.01:max(xmuset);
		plot(xmuset,-log(psimuset),'.',xgrid,-log(wpsi'*xgrid+thetapsi),'r');
	case 'J'
		x1=x-dx;
		dmu=sqrt((x1-xmu).^2);
		[dmusort dmui]=sort(dmu);
		muset=dmui(1:m);
		xmuset1=xmu(muset);
		logpsimuset1=logpsimu(muset);
		if 1
			J1=-lambda*log(mean(exp(logpsimuset1)));
		else
			J1=-lambda*max(logpsimuset1);
		end;
		x2=x+dx;
		dmu=sqrt((x2-xmu).^2);
		[dmusort dmui]=sort(dmu);
		muset=dmui(1:m);
		xmuset2=xmu(muset);
		logpsimuset2=logpsimu(muset);
		if 1
			J2=-lambda*log(mean(exp(logpsimuset2)));
		else
			J2=-lambda*max(logpsimuset2);
		end;
		u=-1/R*(J2-J1)/dx/2;
		%u=0;
		if 0
			subplot(2,2,4)
			plot(xmuset1,-lambda*logpsimuset1,'.',xmuset2,-lambda*logpsimuset2,'.',x1,J1,'*r',x2,J2,'*r',x1,0.5*P*x1^2,'*k',x2,0.5*P*x2^2,'*k')
			[x u -P/R*x]
			drawnow
			pause
		end;
	case 'dxmu'
		% find m nearest to x+dx
		dmu=sqrt((x-xmu).^2);
		[dmusort dmui]=sort(dmu);
		muset=dmui(1:m);
		psimuset=exp(logpsimu(muset));
		ximuset=ximu(muset);
		xmuset=xmu(muset);
		u=psimuset*ximuset'/sum(psimuset)/dt;
		if j==N1,
			subplot(4,2,5)
			hist(xmuset,100)
			xlabel('x');
			subplot(4,2,6)
			hist(ximuset,100)
			xlabel('\xi');
			subplot(4,2,7)
			hist(psimuset,100)
			xlabel('\psi');
			subplot(4,2,8)
			hist(logpsimu(muset),100)
			xlabel('log \psi');
			[x, u, -P*x]
			pause
		end;
		deltax(j)=std(xmuset);
		meanx(j)=mean(xmuset);
		if (j==1)
			fprintf('estimate of optimal cost = %f\n',-lambda*log(mean(psimuset)));
		end;
	end;
	%[x u -P*x]
	%pause
   	%x=x+u*dt+sqrt(nu*dt)*randn;
	x=x+u*dt;

	Vall(j)=0.5*Q*x^2;
	uall(j)=u;
end;
subplot(4,2,2)
plot(dt:dt:Ttotal-T,xcontrol);
xlabel('t');
ylabel('x');
subplot(4,2,3)
plot(dt:dt:Ttotal-T,uall,dt:dt:Ttotal-T,-P/R*xcontrol,'r')
legend('computed','exact')
xlabel('t');
ylabel('u');
subplot(4,2,4)
plot(dt:dt:Ttotal-T,meanx,dt:dt:Ttotal-T,meanx+deltax,dt:dt:Ttotal-T,meanx-deltax);
fprintf('total realized cost in 0 to T = %f\n',sum(Vall(1:T/dt))*dt+0.5*R*sum(uall(1:T/dt).^2)*dt);
fprintf('optimal deterministic cost in 0 to T = %f\n',0.5*P*x0^2)

