% optimal control computation in quadratic potential 
% cost is int_0^T 0.5 R u^2 + 0.5 Q x^2
% dynamics is dx=d\xi
% detailed study of effect of number of samples and nearby data points on estimate
% of u
clear all

rand('state',0);
randn('state',0);

R=1;
Q=5;
nu=1;
T=1;		% horizon time
kburn=30;
Tburn=5;
dt=0.01;
N=T/dt;		% number of time points
Nburn=Tburn/dt;		% number of time points
N1=kburn*(Nburn-N+1)+N;	
Nimp=10000;
Nlarge=N1+Nimp;		% number of time points
Nwindow=1000;
lambda=R*nu;
sigma=sqrt(nu*dt);
P=sqrt(Q*R)*tanh(sqrt(Q/R)*T);
ssmin=500;

dx0=0.2;
xburn=zeros(kburn,Nburn+1);
Svburn=zeros(kburn,Nburn);
wburn=zeros(kburn,Nburn-N+1);
xiburn=sigma*randn(kburn,Nburn);

l_iset=zeros(1,Nlarge);
ss=zeros(1,Nlarge);
u=zeros(1,Nlarge);
u0=zeros(1,Nlarge);
x=zeros(1,Nlarge);
w=zeros(1,Nlarge);
Sv=zeros(1,Nlarge);
Sk=zeros(1,Nlarge);
xi=sigma*randn(1,Nlarge);	
%xi(Nburn+1:Nlarge)=0;
dx1all=zeros(1,Nlarge);

for k=1:kburn,
	xburn(k,1)=2;
	for i=1:Nburn,
		xburn(k,i+1)=xburn(k,i)+xiburn(k,i);			% xburn(k,1:Nburn+1)
		Svburn(k,i)=0.5*Q*dt*xburn(k,i)^2;				% Svburn(k,1:Nburn)
	end;
	for i=1:Nburn-N+1,
		wburn(k,i)=exp(-sum(Svburn(k,i:i+N-1))/lambda); 	% wburn(k,1:Nburn-N+1)
	end;
end;
x=reshape(xburn(1:kburn,1:Nburn-N+1)',1,kburn*(Nburn-N+1));
w=reshape(wburn(1:kburn,1:Nburn-N+1)',1,kburn*(Nburn-N+1));		
N1=kburn*(Nburn-N+1)+N;					% x(1:N1), w(1:N1-N), xi(1:Nlarge)
x(1:N1)=[x,xburn(kburn,Nburn-N+2:Nburn+1)];
xi(1:N1-N)=reshape(xiburn(1:kburn,1:Nburn-N+1)',1,kburn*(Nburn-N+1));		


for i=N1:Nlarge,
	% compute set of points near x(i)
	% w(1:i-N) exist
	%range=i-N:-10:1;
	%range=max(i-10000,1):i-N;
	range=1:i-N;
	xselect=x(range);
	distance=abs(x(i)-xselect);
	iset=find(distance<dx0);
	w_iset=w(range(iset));
	alpha=w_iset/sum(w_iset);
	ss(i)=1/sum(alpha.^2);
	if isinf(ss(i)), ss(i)=0;end;
	l_iset(i)=length(iset);
	dx1=dx0;
	while (ss(i)<ssmin)&(l_iset(i)<i-N)&(dx1<1)
		dx1=dx1+dx0;
		iset=find(distance<dx1);
		l_iset(i)=length(iset);
		w_iset=w(range(iset));
		alpha=w_iset/sum(w_iset);
		ss(i)=1/sum(alpha.^2);
		if isinf(ss(i)), ss(i)=0;end;
	end;
	dx1all(i)=dx1;
	x_iset=x(range(iset));
	xi_iset=xi(range(iset));
	u_iset=u(range(iset));

	% compute u(i)
	if ss(i)<ssmin, 
		u(i)=0;
		disp('u=0')
		i, ss(i)
	else
		barx=alpha*x_iset';
    	chi=alpha*((x_iset-barx).^2)';
	
		% local model for dxi
    	barxi=alpha*xi_iset';
    	etaxi=alpha*((xi_iset-barxi).*(x_iset-barx))';
    	wxi=etaxi*inv(chi);
    	dxi=wxi*(x(i)-barx)+barxi;

		% local model for u0
    	baru=alpha*u_iset';
    	etau=alpha*((u_iset-baru).*(x_iset-barx))';
    	wu=etau*inv(chi);
		u0(i)=wu*(x(i)-barx)+baru;
	
		u(i)=dxi/dt+u0(i);
		if abs(u(i)+P/R*x(i))>5, i,[u(i), -P/R*x(i)], pause, end;
	end;
	error(i)=u(i)+P/R*x(i);

	% compute x(i+1)
	x(i+1)=x(i)+u(i)*dt+xi(i);


	% compute Sv(i), Sk(i)
	Sv(i)=0.5*Q*dt*x(i)^2;
	Sk(i)=0.5*R*dt*u(i)^2+R*xi(i)*u(i);

	% compute w(i-N+1)
	w(i-N+1)=exp(-sum(Sv(i-N+1:i)+Sk(i-N+1:i))/lambda); 	% w(1:i-N+1)
end;
figure(1)
clf
for j=1:(Nlarge-N1)/Nwindow,
	subplot(ceil((Nlarge-N1)/Nwindow/4),4,j)
	range=N1+(j-1)*Nwindow:N1+j*Nwindow-1;
	plot(x(range),u(range),'.',x(range),u0(range),'.r',x(range),-P/R*x(range),'k')
	xlabel('x')
	ylabel('u')
end;
figure(2)
clf
subplot(2,2,1)
plot(N1:Nlarge,ss(N1:Nlarge))
ylabel('sample size');
subplot(2,2,2)
plot(N1:Nlarge,l_iset(N1:Nlarge))
ylabel('length iset');
subplot(2,2,3)
plot(N1:Nlarge,dx1all(N1:Nlarge))
ylabel('dx1');
subplot(2,2,4)
plot(N1:Nlarge,error(N1:Nlarge))
ylabel('error');
figure(3)
clf
plot(x)
hold on
plot(u,'r')
plot(N1*ones(1,9),(-4:4),'k')
hold off
